
/* Chrysalide - Outil d'analyse de fichiers binaires
 * elf.c - support du format ELF
 *
 * Copyright (C) 2009-2013 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  OpenIDA is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  OpenIDA is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "elf.h"


#include <malloc.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>


#include <i18n.h>


#include "elf-int.h"
#include "program.h"
#include "section.h"
#include "strings.h"
#include "symbols.h"
#include "../../gui/panels/log.h"






/* Taille maximale d'une description */
#define MAX_PORTION_DESC 256



/* Initialise la classe des formats d'exécutables ELF. */
static void g_elf_format_class_init(GElfFormatClass *);

/* Initialise une instance de format d'exécutable ELF. */
static void g_elf_format_init(GElfFormat *);

/* Supprime toutes les références externes. */
static void g_elf_format_dispose(GElfFormat *);

/* Procède à la libération totale de la mémoire. */
static void g_elf_format_finalize(GElfFormat *);

/* Informe quant au boutisme utilisé. */
static SourceEndian g_elf_format_get_endianness(const GElfFormat *);

/* Indique le type d'architecture visée par le format. */
static const char *g_elf_format_get_target_machine(const GElfFormat *);

/* Etend la définition des portions au sein d'un binaire. */
static void g_elf_format_refine_portions(const GElfFormat *, GPortionLayer *);

/* Fournit l'emplacement correspondant à une position physique. */
static bool g_elf_format_translate_offset_into_vmpa(const GElfFormat *, phys_t, vmpa2t *);

/* Fournit l'emplacement correspondant à une position physique. */
static bool g_elf_format_translate_address_into_vmpa(const GElfFormat *, virt_t, vmpa2t *);

/* Fournit l'emplacement d'une section donnée. */
static bool g_elf_format_get_section_range_by_name(const GElfFormat *, const char *, mrange_t *);



/******************************************************************************
*                                                                             *
*  Paramètres  : content = contenu binaire à parcourir.                       *
*                parent  = éventuel format exécutable déjà chargé.            *
*                unused  = adresse non utilisée ici.                          *
*                key     = identifiant de format trouvé ou NULL. [OUT]        *
*                                                                             *
*  Description : Indique si le format peut être pris en charge ici.           *
*                                                                             *
*  Retour      : Conclusion de haut niveau sur la reconnaissance effectuée.   *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

FormatMatchStatus elf_is_matching(GBinContent *content, GExeFormat *parent, void *unused, char **key)
{
    FormatMatchStatus result;               /* Bilan à renvoyer            */
    vmpa2t addr;                            /* Tête de lecture initiale    */
    bool status;                            /* Bilan des accès mémoire     */
    char magic[4];                          /* Idenfiant standard          */

    if (parent != NULL)
        return FMS_UNKNOWN;

    init_vmpa(&addr, 0, VMPA_NO_VIRTUAL);

    status = g_binary_content_read_raw(content, &addr, 4, (bin_t *)magic);

    status &= (memcmp(magic, "\x7f\x45\x4c\x46" /* .ELF */, 4) == 0);

    if (status)
    {
        result = FMS_MATCHED;
        *key = strdup("elf");
    }
    else
        result = FMS_UNKNOWN;

    return result;

}


/* Indique le type défini pour un format d'exécutable ELF. */
G_DEFINE_TYPE(GElfFormat, g_elf_format, G_TYPE_EXE_FORMAT);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des formats d'exécutables ELF.          *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_elf_format_class_init(GElfFormatClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GBinFormatClass *fmt;                   /* Version en format basique   */
    GExeFormatClass *exe;                   /* Version en exécutable       */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_elf_format_dispose;
    object->finalize = (GObjectFinalizeFunc)g_elf_format_finalize;

    fmt = G_BIN_FORMAT_CLASS(klass);

    fmt->get_endian = (format_get_endian_fc)g_elf_format_get_endianness;

    exe = G_EXE_FORMAT_CLASS(klass);

    exe->get_machine = (get_target_machine_fc)g_elf_format_get_target_machine;
    exe->refine_portions = (refine_portions_fc)g_elf_format_refine_portions;

    exe->translate_phys = (translate_phys_fc)g_elf_format_translate_offset_into_vmpa;
    exe->translate_virt = (translate_virt_fc)g_elf_format_translate_address_into_vmpa;

    exe->get_range_by_name = (get_range_by_name_fc)g_elf_format_get_section_range_by_name;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : format = instance à initialiser.                             *
*                                                                             *
*  Description : Initialise une instance de format d'exécutable ELF.          *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_elf_format_init(GElfFormat *format)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : format = instance d'objet GLib à traiter.                    *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_elf_format_dispose(GElfFormat *format)
{
    G_OBJECT_CLASS(g_elf_format_parent_class)->dispose(G_OBJECT(format));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : format = instance d'objet GLib à traiter.                    *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_elf_format_finalize(GElfFormat *format)
{
    G_OBJECT_CLASS(g_elf_format_parent_class)->finalize(G_OBJECT(format));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : content = contenu binaire à parcourir.                       *
*                parent  = éventuel format exécutable déjà chargé.            *
                 status  = barre de statut à tenir informée.                  *
*                                                                             *
*  Description : Prend en charge un nouveau format ELF.                       *
*                                                                             *
*  Retour      : Adresse de la structure mise en place ou NULL en cas d'échec.*
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GBinFormat *g_elf_format_new(GBinContent *content, GExeFormat *parent, GtkStatusStack *status)
{
    GElfFormat *result;                     /* Structure à retourner       */

    result = g_object_new(G_TYPE_ELF_FORMAT, NULL);

    g_binary_format_set_content(G_BIN_FORMAT(result), content);

    if (!read_elf_header(result, &result->header, &result->is_32b, &result->endian))
    {
        /* TODO */
        return NULL;
    }


    /* Vérification des tailles d'entrée de table */

    if (ELF_HDR(result, result->header, e_phentsize) != ELF_SIZEOF_PHDR(result))
    {
        log_variadic_message(LMT_BAD_BINARY, _("Corrupted program header size (%hu); fixed !  --  replacing 0x%04hx by 0x%04hx at offset 0x%x"),
                             ELF_HDR(result, result->header, e_phentsize),
                             ELF_HDR(result, result->header, e_phentsize),
                             ELF_SIZEOF_PHDR(result), ELF_HDR_OFFSET_OF(result, e_phentsize));
        ELF_HDR_SET(result, result->header, e_phentsize, ELF_SIZEOF_PHDR(result));
    }

    if (ELF_HDR(result, result->header, e_shentsize) != ELF_SIZEOF_SHDR(result))
    {
        log_variadic_message(LMT_BAD_BINARY, _("Corrupted section header size (%hu); fixed !  --  replacing 0x%04hx by 0x%04hx at offset 0x%x"),
                             ELF_HDR(result, result->header, e_shentsize),
                             ELF_HDR(result, result->header, e_shentsize),
                             ELF_SIZEOF_SHDR(result), ELF_HDR_OFFSET_OF(result, e_shentsize));
        ELF_HDR_SET(result, result->header, e_shentsize, ELF_SIZEOF_SHDR(result));
    }

    /* FIXME : à améliorer */
    /*
    if ((ELF_HDR(result, result->header, e_shnum) * ELF_HDR(result, result->header, e_shentsize)) >= length)
    {
        log_variadic_message(LMT_BAD_BINARY, ("Suspicious section table (bigger than the binary !) ; reset !  --  replacing 0x%04hx by 0x%04hx at offset 0x%x"),
                             ELF_HDR(result, result->header, e_shnum),
                             0, ELF_HDR_OFFSET_OF(result, e_shnum));
        ELF_HDR_SET(result, result->header, e_shnum, 0);
    }
    */



    if (!load_elf_symbols(result, status))
    {
        /* TODO */
        return NULL;
    }


    if (!find_all_elf_strings(result))
    {
        /* TODO */
        return NULL;
    }


    if (!g_binary_format_complete_loading(G_BIN_FORMAT(result)))
        goto gefn_error;

    return G_BIN_FORMAT(result);

 gefn_error:

    g_object_unref(G_OBJECT(result));

    return NULL;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : format = informations chargées à consulter.                  *
*                                                                             *
*  Description : Informe quant au boutisme utilisé.                           *
*                                                                             *
*  Retour      : Indicateur de boutisme.                                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static SourceEndian g_elf_format_get_endianness(const GElfFormat *format)
{
    return format->endian;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : format = informations chargées à consulter.                  *
*                                                                             *
*  Description : Indique le type d'architecture visée par le format.          *
*                                                                             *
*  Retour      : Identifiant de l'architecture ciblée par le format.          *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static const char *g_elf_format_get_target_machine(const GElfFormat *format)
{
    const char *result;                     /* Identifiant à retourner     */

    switch (ELF_HDR(format, format->header, e_machine))
    {
        case EM_386:
            result = "i386";
            break;

        case EM_MIPS:
            result = "mips";
            break;

        case EM_ARM:
            result = "armv7";
            break;

        case EM_NONE:
        default:
            result = NULL;
            break;

    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : format = informations chargées à consulter.                  *
*                main   = couche de portions principale à raffiner.           *
*                                                                             *
*  Description : Etend la définition des portions au sein d'un binaire.       *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_elf_format_refine_portions(const GElfFormat *format, GPortionLayer *main)
{
    GPortionLayer *layer;                   /* Couche à mettre en place    */
    uint16_t max;                           /* Décompte d'éléments traités */
    elf_phdr *sorted_phdrs;                 /* Liste de segments triée     */
    uint16_t i;                             /* Boucle de parcours          */
    off_t offset;                           /* Début de part de programme  */
    elf_phdr *phdr;                         /* En-tête de programme ELF    */
    uint32_t p_flags;                       /* Droits associés à une partie*/
    const char *background;                 /* Fond signigicatif           */
    GBinPortion *new;                       /* Nouvelle portion définie    */
    char desc[MAX_PORTION_DESC];            /* Description d'une portion   */
    vmpa2t addr;                            /* Emplacement dans le binaire */
    PortionAccessRights rights;             /* Droits d'une portion        */
    elf_shdr strings;                       /* Section des descriptions    */
    bool has_strings;                       /* Section trouvée ?           */
    elf_shdr *sorted_shdrs;                 /* Liste de sections triée     */
    elf_shdr *section;                      /* En-tête de section ELF      */
    uint64_t sh_flags;                      /* Droits associés à une partie*/
    const char *name;                       /* Nom trouvé ou NULL          */

    /**
     * La copie des différents en-têtes cherche à reproduire l'inclusion native
     * du format :
     *
     *  EXIDX          0x001178 0x00009178 0x00009178 0x00008 0x00008 R   0x4
     *  PHDR           0x000034 0x00008034 0x00008034 0x00120 0x00120 R E 0x4
     *  INTERP         0x000154 0x00008154 0x00008154 0x00019 0x00019 R   0x1
     *  LOAD           0x000000 0x00008000 0x00008000 0x01184 0x01184 R E 0x8000
     *
     */

    /**
     * Côté segments basiques.
     */

    layer = g_portion_layer_new(NO_LENGTH_YET, _("Segment"));

    g_portion_layer_attach_sub(main, layer);

    /* Constitution d'une liste de travail */

    max = ELF_HDR(format, format->header, e_phnum);

    sorted_phdrs = (elf_phdr *)calloc(max, sizeof(elf_phdr));

    for (i = 0; i < max; i++)
    {
        offset = ELF_HDR(format, format->header, e_phoff)
            + ELF_HDR(format, format->header, e_phentsize) * i;

        if (!read_elf_program_header(format, offset, &sorted_phdrs[i]))
        {
            if (format->is_32b)
                sorted_phdrs[i].phdr32.p_type = PT_NULL;
            else
                sorted_phdrs[i].phdr64.p_type = PT_NULL;
        }

    }

    /* Tri de cette liste */

    int sort_phdr(elf_phdr *a, elf_phdr *b)
    {
        uint64_t filesz_a;                  /* Taille de l'en-tête 'a'     */
        uint64_t filesz_b;                  /* Taille de l'en-tête 'b'     */
        int status;                         /* Bilan d'une comparaison     */

        filesz_a = ELF_PHDR(format, *a, p_filesz);
        filesz_b = ELF_PHDR(format, *b, p_filesz);

        if (filesz_a < filesz_b)
            status = 1;

        else if (filesz_a > filesz_b)
            status = -1;

        else
            status = 0;

        return status;

    }

    qsort(sorted_phdrs, max, sizeof(elf_phdr), (__compar_fn_t)sort_phdr);

    /* Inclusion de ces en-têtes */

    for (i = 0; i < max; i++)
    {
        phdr = &sorted_phdrs[i];

        if (ELF_PHDR(format, *phdr, p_type) == PT_NULL)
            continue;

        p_flags = ELF_PHDR(format, *phdr, p_flags);

        if (p_flags & PF_X) background = BPC_CODE;
        else if (p_flags & PF_W) background = BPC_DATA;
        else background = BPC_DATA_RO;

        new = g_binary_portion_new(background);

        snprintf(desc, MAX_PORTION_DESC, "%s \"%s\"",
                 _("Segment"),
                 get_elf_program_type_desc(ELF_PHDR(format, *phdr, p_type)));

        g_binary_portion_set_desc(new, desc);

        init_vmpa(&addr, ELF_PHDR(format, *phdr, p_offset), ELF_PHDR(format, *phdr, p_vaddr));
        g_binary_portion_set_values(new, &addr, ELF_PHDR(format, *phdr, p_filesz));

        rights = PAC_NONE;
        if (p_flags & PF_R) rights |= PAC_READ;
        if (p_flags & PF_W) rights |= PAC_WRITE;
        if (p_flags & PF_X) rights |= PAC_EXEC;

        g_binary_portion_set_rights(new, rights);

        g_portion_layer_include(layer, new);

    }

    free(sorted_phdrs);

    /**
     * Inclusion des sections, si possible...
     */

    has_strings = find_elf_section_by_index(format,
                                            ELF_HDR(format, format->header, e_shstrndx),
                                            &strings);

    layer = g_portion_layer_new(NO_LENGTH_YET, _("Section"));

    g_portion_layer_attach_sub(main, layer);

    /* Constitution d'une liste de travail */

    max = ELF_HDR(format, format->header, e_shnum);

    sorted_shdrs = (elf_shdr *)calloc(max, sizeof(elf_shdr));

    for (i = 0; i < max; i++)
    {
        if (!find_elf_section_by_index(format, i, &sorted_shdrs[i]))
        {
            if (format->is_32b)
                sorted_shdrs[i].shdr32.sh_offset = 0;
            else
                sorted_shdrs[i].shdr64.sh_offset = 0;
        }

    }

    /* Tri de cette liste */

    int sort_shdr(elf_shdr *a, elf_shdr *b)
    {
        uint64_t size_a;                    /* Taille de l'en-tête 'a'     */
        uint64_t size_b;                    /* Taille de l'en-tête 'b'     */
        int status;                         /* Bilan d'une comparaison     */

        size_a = ELF_SHDR(format, *a, sh_size);
        size_b = ELF_SHDR(format, *b, sh_size);

        if (size_a < size_b)
            status = 1;

        else if (size_a > size_b)
            status = -1;

        else
            status = 0;

        return status;

    }

    qsort(sorted_shdrs, max, sizeof(elf_shdr), (__compar_fn_t)sort_shdr);

    /* Inclusion de ces en-têtes */

    for (i = 0; i < max; i++)
    {
        section = &sorted_shdrs[i];

        if (ELF_SHDR(format, *section, sh_offset) == 0)
            continue;

        sh_flags = ELF_SHDR(format, *section, sh_flags);

        if ((sh_flags & SHF_ALLOC) == 0)
            continue;

        if (sh_flags & SHF_EXECINSTR) background = BPC_CODE;
        else if (sh_flags & SHF_WRITE) background = BPC_DATA;
        else background = BPC_DATA_RO;

        new = g_binary_portion_new(background);

        if (has_strings)
            name = extract_name_from_elf_string_section(format, &strings,
                                                        ELF_SHDR(format, *section, sh_name));
        else name = NULL;

        if (name != NULL)
            sprintf(desc, "%s \"%s\"", _("Section"), name);
        else
            sprintf(desc, "%s ???", _("Section"));

        g_binary_portion_set_desc(new, desc);

        init_vmpa(&addr, ELF_SHDR(format, *section, sh_offset), ELF_SHDR(format, *section, sh_addr));
        g_binary_portion_set_values(new, &addr, ELF_SHDR(format, *section, sh_size));

        rights = PAC_NONE;
        if (sh_flags & SHF_ALLOC) rights |= PAC_READ;
        if (sh_flags & SHF_WRITE) rights |= PAC_WRITE;
        if (sh_flags & SHF_EXECINSTR) rights |= PAC_EXEC;

        g_binary_portion_set_rights(new, rights);

        g_portion_layer_include(layer, new);

    }

    free(sorted_shdrs);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : format = description de l'exécutable à consulter.            *
*                off    = position physique à retrouver.                      *
*                pos    = position correspondante. [OUT]                      *
*                                                                             *
*  Description : Fournit l'emplacement correspondant à une position physique. *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_elf_format_translate_offset_into_vmpa(const GElfFormat *format, phys_t off, vmpa2t *pos)
{
    bool result;                            /* Bilan à retourner           */

    result = translate_offset_into_vmpa_using_elf_sections(format, off, pos);

    if (!result)
        result = translate_offset_into_vmpa_using_elf_programs(format, off, pos);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : format = description de l'exécutable à consulter.            *
*                addr   = adresse virtuelle à retrouver.                      *
*                pos    = position correspondante. [OUT]                      *
*                                                                             *
*  Description : Fournit l'emplacement correspondant à une adresse virtuelle. *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_elf_format_translate_address_into_vmpa(const GElfFormat *format, virt_t addr, vmpa2t *pos)
{
    bool result;                            /* Bilan à retourner           */

    result = translate_address_into_vmpa_using_elf_sections(format, addr, pos);

    if (!result)
        result = translate_address_into_vmpa_using_elf_programs(format, addr, pos);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : format = description de l'exécutable à consulter.            *
*                name   = nom de la section recherchée.                       *
*                range  = emplacement en mémoire à renseigner. [OUT]          *
*                                                                             *
*  Description : Fournit l'emplacement d'une section donnée.                  *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_elf_format_get_section_range_by_name(const GElfFormat *format, const char *name, mrange_t *range)
{
    bool result;                            /* Bilan à retourner           */
    phys_t offset;                          /* Position physique de section*/
    phys_t size;                            /* Taille de la section trouvée*/
    virt_t address;                         /* Adresse virtuelle de section*/
    vmpa2t tmp;                             /* Adresse à initialiser       */

    result = find_elf_section_content_by_name(format, name, &offset, &size, &address);

    if (result)
    {
        init_vmpa(&tmp, offset, address);
        init_mrange(range, &tmp, size);
    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : format = informations chargées à consulter.                  *
*                                                                             *
*  Description : Présente l'en-tête ELF du format chargé.                     *
*                                                                             *
*  Retour      : Pointeur vers la description principale.                     *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

const elf_header *g_elf_format_get_header(const GElfFormat *format)
{
    return &format->header;

}
